# frozen_string_literal: true

require 'ephemeron/railtie'

require 'ephemeron/addons/controller_addons'
require 'ephemeron/addons/model_addons'

require 'ephemeron/writers/thread_store'
require 'ephemeron/writers/callbacks'
require 'ephemeron/writers/logger'
require 'ephemeron/writers/store'

require 'ephemeron/errors'
require 'ephemeron/logger/convert'
require 'ephemeron/logger/finders'
require 'ephemeron/settings'
require 'ephemeron/store/allow_fetch'
require 'ephemeron/store/convert'
require 'ephemeron/store/finders'
require 'ephemeron/store/update'
require 'ephemeron/validators/before_save'
require 'ephemeron/validators/store'
require 'ephemeron/validators/store_element'
require 'ephemeron/validators/store/allowed_but_unsaved'
require 'ephemeron/validators/store/unused'

module Ephemeron
  module_function

  def configure
    Ephemeron::Settings::CONFIG.new.tap do |config|
      yield config
      Ephemeron::Settings.configure config
    end
  end

  def state
    Store.store
  end

  def logs
    Logger.logs
  end

  def add(model, as: nil)
    Store.add model, as: as
  end

  def proxy
    Logger.log nil, :proxy_start
    res = yield
    Logger.log nil, :proxy_end
    res
  end

  def get(key)
    Store.get key
  end

  def used(model_s)
    log_used = ->(model) { Logger.log model, :used }
    if model_s.is_a? ActiveRecord::Base
      log_used.call model_s
    else
      model_s.each(&log_used)
    end
    model_s
  end

  def allow_fetch!(model_or_key)
    model = if model_or_key.is_a?(String)
              Ephemeron.get(model_or_key)
            else
              model_or_key
            end
    model.tap { |m| Logger.log m, :fetch_allowed }
  end

  def allow_save!(model)
    model.tap { |m| Logger.log m, :save_allowed }
  end

  def skip_save!(model)
    Logger.log model, :skip_save
  end

  def after_save!(&block)
    Callbacks.add block
  end

  def finish
    if Store.aliases[nil] != :abort
      Store.validate
      Store.persist
      Callbacks.call
    end
    reset
  end

  def stop
    Store.aliases[nil] = :abort
  end

  def reset
    Store.reset
    Logger.reset
    Callbacks.reset
    true
  end
end
