# frozen_string_literal: true

module Ephemeron
  class Store < ThreadStore
    attribute :ts_store, :ts_aliases

    def store
      self.ts_store ||= {}
    end

    def aliases
      self.ts_aliases ||= {}
    end

    def has?(model, as_new_record: false)
      !store[Convert.call(model, as_new_record: as_new_record)].nil?
    end

    def add(model, opts = {})
      return unless Validators::StoreElement.call(model)

      in_store_before = stored?(model)
      store[to_store_key(model)] = model unless stored?(model)
      Logger.log(model, :fetched) if opts[:log]
      add_alias(model, opts[:as], in_store_before)
      model
    end

    def get(key)
      model = store[key]
      model ||= store[aliases[key]]
      Logger.log(model, :used) if model
      model
    end

    def validate
      Validators::Store.call
    end

    def persist
      changed_models = store.filter { |_, v| v.changed? }.values
                            .filter { |model| !Logger.has?(model, :skip_save) }
      ActiveRecord::Base.transaction do
        changed_models.each do |model|
          Ephemeron.allow_save! model
          model.save!
        end
      end
    end

    def destroy(model)
      key = Convert.call model
      return if store[key].nil?

      store.delete key
    end

    def reset
      self.ts_store = {}
      self.ts_aliases = {}
    end

    private

    def to_store_key(model)
      Convert.call(model)
    end

    def stored?(model)
      !store[to_store_key(model)].nil?
    end

    def add_alias(model, as, in_store_before)
      if !as.nil?
        aliases[as] = to_store_key(model)
      elsif in_store_before && !Store::AllowFetch.call(model)
        raise StoreError, "#{to_store_key(model)} is already in the store"
      end
    end
  end
end
